﻿using System;
using System.Collections.Generic;

using System.Linq;

static class EnumerableExtensions {
	public static void PrintAll<T>(this IEnumerable<T> list) {
		foreach (var item in list) {
			Console.Write(item);
			Console.Write(' ');
		}
		Console.WriteLine();
	}
}

static class LinkedListExtensions {
	public static void Add<T>(this LinkedList<T> list, T item)
		=> list.AddLast(item);

	public static IEnumerable<LinkedListNode<T>> AsNodeEnumerable<T>(this LinkedList<T> list) {
		var current = list.First;
		while (current != null) {
			yield return current;
			current = current.Next;
		}
	}
}

class Program {
	static void Main(string[] args) {
		var list = new LinkedList<int> { 1, 2, 3, 4, 5, 6 };    // Works thanks to Add<T>(this LinkedList<T> list, T item) extension method above!

		Console.WriteLine("Original: "); list.PrintAll();

		// Lazy evaluation
		foreach (var node in list.AsNodeEnumerable()) {
			node.Value *= -1;
		}

		Console.WriteLine("After lazy update: "); list.PrintAll();

		// Lazy evaluation
		foreach (var node in list.AsNodeEnumerable()) {
			if (node.Value % 2 == 0) {
				node.List.AddBefore(node, node.Value * -1);			// -> ADD HERE -> node ->
			}
		}

		Console.WriteLine("After lazy AddBefore: "); list.PrintAll();

		// Eager evaluation, so it does a full list copy !!!
		foreach (var node in list.AsNodeEnumerable().ToList()) {	// ToList() from System.Linq
			if (node.Value % 2 == 0) {
				node.List.AddAfter(node, node.Value * 100);			// -> node -> ADD HERE ->
			}
		}

		Console.WriteLine("After eager AddAfter: "); list.PrintAll();

        #region Lazy unwound
        // Lazy evaluation, a bit more complex, but works fine
        var e = list.AsNodeEnumerable().GetEnumerator();
		bool hasItems = e.MoveNext();
		while (hasItems) {
			var node = e.Current;
			hasItems = e.MoveNext();
            if (node.Value % 2 == 0) {
                node.List.AddAfter(node, node.Value * 100);         // -> node -> ADD HERE ->
            }
        }
		e.Dispose();

        Console.WriteLine("After unwound lazy AddAfter: "); list.PrintAll();
        #endregion

        // Lazy evaluation, but will eat all available memory !!!
        foreach (var node in list.AsNodeEnumerable()) {
			if (node.Value % 2 == 0) {
				node.List.AddAfter(node, node.Value * 100);			// -> node -> ADD HERE ->
			}
		}

		Console.WriteLine("After lazy AddAfter: "); list.PrintAll();
	}
}


